
################################################################
# $Id: MKalendar.pm,v 1.3 2002/10/22 07:02:20 martink Exp $
################################################################
#
# CVS Revision: $Revision: 1.3 $
# Commit date:  $Date: 2002/10/22 07:02:20 $
#         Tag:  $Name:  $
#      Author:  $Author: martink $
#
################################################################

################################################################
# 
# Very funky web based calendar object. Uses the Date::Calc module
# for its main date routines.
#
################################################################

package MKalendar;

use Carp;
use CGI qw/:all/;
use POSIX qw(strftime floor ceil);
use Date::Parse;
use Date::Calc qw(Today Days_in_Month 
		  Month_to_Text
		  Add_Delta_Days
		  Day_of_Week_to_Text
		  Week_Number
		  Decode_Day_of_Week
		  Now
		  Day_of_Week
		  Delta_Days);
use Time::HiRes;
use vars qw($AUTOLOAD $VERSION $CVSTAG $CVSDATE $DEBUG $NOWYEAR $NOWMONTH $NOWDAY $EXECSTART $MODIFIERS $HTML);

use strict;

$DEBUG = 0; 

################################################################
# CVS headers and variables: VERSION(=revision), TAG and DATE.
# The date is formatted to the local time zone.
$VERSION = q{ $Revision: 1.3 $ };
$CVSTAG  = q{ $Name:  $ };
$CVSDATE = q{ $Date: 2002/10/22 07:02:20 $ };
$VERSION = $1 if $VERSION =~ /\$.*?:\s*(.*?)\s*\$/;
$CVSDATE = $1 if $CVSDATE =~ /\$Date: (.*?) /;
$CVSTAG  = $1 if $CVSTAG =~ /\$.*?:\s*(.*?)\s*\$/;
################################################################

################################################################
# Some global date things
#
# NOWYEAR NOWMONTH NOWDAY - store today's date
# EXECSTART - stores start of module execution
# MODIFIERS - stores various modifiers for date cells and events
($NOWYEAR,$NOWMONTH,$NOWDAY) = Today();
$EXECSTART = [ Time::HiRes::gettimeofday() ];
$MODIFIERS = [qw(WEEKCOL TODAY OTHERMONTH THISWEEK TODAYCOLUMN WEEKEND WEEKEVENT DAYEVENT MONTHEVENT)];
#
################################################################

sub new {

  my ($class) = shift;
  my %args = @_;

  $EXECSTART = [ Time::HiRes::gettimeofday() ];

  # Definitions of fields
  my $self = {
    _month         => $MKalendar::NOWMONTH,#  
    _year          => $MKalendar::NOWYEAR, #  
    _daylinks      => {}, # links from the day numbers of each cell
    _events        => [], # events of the calendar
    _dayheaders    => [],
    _dayevents     => [],
    _weekevents    => [],
    _HTML          => {},
    _COLOURS       => {},
    _VARS          => {
		       MODIFIER=>{},
		      },
    _monthevents   => [],
    _tableopts     => {},
    _columnopts    => {},
    _calendaropts  => {},
    _colourscheme  => undef,
    _rowopts       => {},
    _cellopts      => {},
    _viewarray     => [],
    _viewheaders   => [],
    _html          => 1,        # html flag
  };
  
  # Set defaults to all option fields
  $self = { %$self,
	   _calendaropts=>{
			   show_version=>1,
			   show_bench=>0,
			  },
	   _tableopts=>{
			sizeopts=>"rel",
			size=>"100%",
#			disabledow=>[qw(sat sun)],
			startdow=>"sun"
		       },
	   _columnopts=>{
			 sizeopts=>"rel",
			 align=>"center",
			 weekcol=>0,
			 weekcolpos=>0,
			},
	   _colourscheme=>"prodblue",
	   _HTML => {
		     "preamble"=>"<br>",
		     "calendar_s"=>"<table width=VARCALWIDTH border=1 cellspacing=0 cellpadding=3>",
		     "calendarheader_init"=>"<tr>",
		     "calendarheader_s"=>"<td colspan=VARNCOLUMN align=center class=CLRHEADER>",
		     "calendarheader"=>"<b>VARMONTHTEXT VARYEAR</b>",
		     "calendarheader_f"=>"</td>",
		     "calendarheader_end"=>"</tr>",
		     "columnheader_init"=>"<tr>",
		     "columnheader_s"=>"<td width=VARCOLUMNWIDTH align=center class=CLRCOLUMNHEADER>",
		     "columnheader"=>"<b>VARDOWTEXT</b>",
		     "columnheader_weekcol"=>"<b>WEEK</b>",
		     "columnheader_f"=>"</td>",
		     "columnheader_end"=>"</tr>",
		     "days_s"=>"",
		     "row_s"=>"<tr>",
		     "row_f"=>"</tr>",
		     "cell_init"=>"<td valign=top class=CLRCELL>",
		     "cell_s_today"=>"<span class=small>&#149; VARDOWTEXTSHORT VARMONTHTEXT VARYEAR <b>VARHOUR:VARMIN:VARSEC</b> &#149; </span><br>",
		     "cell_s"=>"<span class=small>VARDOWTEXTSHORT VARMONTHTEXT VARYEAR</span><br>",
		     "cell_dayheader_s"=>"<span class=darkprodbluetext>",
		     "cell_dayheader"=>"<span class=large><span class=CLRCELLDOM><b>VARDAY</b></span></span> &nbsp;",
		     "cell_dayheader_f"=>"</span><br>",
		     "cell_f"=>"",
		     "cell_end"=>"</td>",

		     "cell_init_weekcol"=>"<td valign=top class=CLRCELL>",
		     "cell_s_weekcol"=>"<b>week VARWEEK</b><br>",
		     "cell_end_weekcol"=>"</td>",

		     "event_init"=>"",
		     "event_s_dayevent"=>"",
		     "event_s_monthevent"=>"<b>MONTH EVENTS</b><br>",
		     "event_s"=>"",
		     "event_f"=>"<br>",
		     "event_end_dayevent"=>"&#164;<br>",
		     "event_end_weekevent"=>"&#149;<br>",
		     "days_f"=>"",
		     "footer_init"=>"<tr>",
		     "footer_s"=>"<td align=left class=small colspan=VARNCOLUMN>",
		     "footer"=>"footer",
		     "footer_f"=>"</td>",
		     "footer_end"=>"</tr>",
		     "stamp_init"=>"<tr>",
		     "stamp_s"=>"<td align=right class=small colspan=VARNCOLUMN>",
		     "stamp"=>"VARBENCH sec | vVARVERSION",
		     "stamp_f"=>"</td>",
		     "stamp_end"=>"</tr>",
		     "calendar_f"=>"</table>",
		     "postamble"=>"<br>",
		    },
	  };

  bless $self,$class;

  # Initalize the class from an input hash
  foreach my $field (keys %args) {
    my $fieldvarname = "_".$field;
    if(defined $self->{$fieldvarname}) {
      $self->{$fieldvarname} = $args{$field};
    }
  }

  # Initialize colours

  return $self;
}

{
  my %_attrs = ();
  sub _accessible {
    my ($self,$attr,$mode) = @_;
    if(defined $_attrs{$attr}) {
      $_attrs{$attr} =~ /$mode/;
    } else {
      1;
    }
  }
}

################################################################
# 
# Initialize the colours for the calendar based on a colour
# scheme. This function can be passed the colourscheme#
#
# use the scheme named SCHEME_NAME
# e.g. InitColours(colourscheme=>SCHEME_NAME);
# use the default scheme, previously set
# e.g. InitColours()
#
# If SCHEME_NAME or the default scheme is not defined below,
# this function does nothing.

sub InitColours {
  my $self = shift;
  my $scheme;
  $scheme = $self->get_colourscheme;
  my %COLOURS =
      (robinhood=>
       {
	header=>"vdarkgreen",
	columnheader=>"mediumorange",
	columnheader_todaycolumn=>"vvlightorange",
	columnheader_weekend=>"vdarkgreen",
	columnheader_weekcol=>"vdarkyellow",
	cell=>"vvlightgreen",
	cell_weekend=>"vvlightred",
	cell_weekcol=>"vvlightyellow",
	cell_thisweek=>"vvlightorange",
	cell_todaycolumn=>"vvlightorange",
	cell_today=>"lightgreen",
	celldom=>"vvlightgreen",
	celldom_weekend=>"vvlightred",
	celldom_thisweek=>"vvlightorange",
	celldom_todaycolumn=>"vvlightorange",
	celldom_today=>"lightgreen",
       },
       prodblue=>
       {
	header=>"darkprodblue",
	columnheader=>"mediumprodblue",
	columnheader_todaycolumn=>"vvlightprodblue",
	columnheader_weekend=>"darkprodblue",
	cell=>"white",
	cell_othermonth=>"vvlightgrey",
	cell_weekend=>"vvvlightprodblue",
	cell_thisweek=>"vvvlightprodblue",
	cell_todaycolumn=>"vvlightprodblue",
	cell_today=>"vlightorange",
	celldom=>"white",
	celldom_othermonth=>"vvlightgrey",
	celldom_weekend=>"",
	celldom_thisweek=>"vvvlightprodblue",
	celldom_todaycolumn=>"vvlightprodblue",
	celldom_today=>"vlightorange"
	},
        bw=>
       {
	header=>"black",
	columnheader=>"darkgrey",
	columnheader_todaycolumn=>"lightred",
	columnheader_weekend=>"darkprodblue",
	cell=>"white",
	cell_othermonth=>"vvlightgrey",
	cell_weekend=>"darkred",
	cell_thisweek=>"vvlightgrey",
	cell_todaycolumn=>"vvlightgreen",
	cell_today=>"vlightgreen",
	celldom=>"white",
	celldom_othermonth=>"vvlightgrey",
	celldom_weekend=>"",
	celldom_thisweek=>"vvlightgrey",
	celldom_todaycolumn=>"vdarkgreytext",
	celldom_today=>"vlightgreen"
	},
       );

  if(defined $COLOURS{$scheme}) {
    my $scheme_colours = $COLOURS{$scheme};
    foreach my $colouritem (keys %{$scheme_colours}) {
      my $colourvalue = $scheme_colours->{$colouritem};
      my $itemkey = uc "CLR$colouritem";
      $self->get_VARS->{$itemkey} = $colourvalue;
    }
  }
}
# InitColours
################################################################

################################################################
#
# Replace some variables within the event text with current calendar variables
# VARYEAR        = current year (e.g. 2001)
# VARMONTH       = current month (e.g. 3)
# VARMONTHTEXT   = current month (e.g. March)
# VARWEEKOFMONTH = index of the week of the day in the month (e.g. 2nd week)
# VARDAY         = current day of month (e.g 13)
# VARDOW         = day of week number (e.g. 3)
# VARDOWTEXT     = text of day of week (e.g. Tuesday)
# VARDOWNTH      = index of day of week (e.g. 2 for 2nd tue of month)
# VARCOLUMN      = column in calendar
#
sub ParseText {
  my $self = shift;
  my %input = @_;
  my $text  = $input{text};
  my $param_hash = $input{param};

  # First find all variable occurrences in the text. The variables are one-word strings
  # that start with VAR.
  my @varlist = ();
  my @clrlist = ();
  while($text =~ /(VAR|CLR)([A-Z]+)/g) {
    my $type = $1;
    my $tag = $2;
    if($type eq "VAR") {
      push(@varlist,"VAR".$tag);
    } else {
      push(@clrlist,"CLR".$tag);
    }
  }
  # Now replace all the vars, starting with the longest first. The values to use
  # for the variables are stored in %$param_hash.
  #
  # $param_hash->{VARvariable} = VARvalue.
  my $modifier_hash = $param_hash->{MODIFIER};
  foreach my $tag (sort {length($b) <=> length($a)} (@varlist,@clrlist)) {
    my $value;
    foreach my $modifier (@{$MKalendar::MODIFIERS}) {
      if($modifier_hash->{$modifier}) {
	$value = $param_hash->{"${tag}_$modifier"};
	last;
      }
    }
    if(! defined $value) {
      $value = $param_hash->{$tag};
    }
    if(defined $value) {
      $text =~ s/$tag/$value/g;
    }
  }
  return $text;
}
# ParseText
################################################################

################################################################
# Parse the text within an event object: text and link.
#
# ParseEvent(event=>$event_reference,
#            param=>$parameters_reference);
sub ParseEvent {
  my $self = shift;
  my %input = @_;
  my $event_ref = $input{event} || return;
  my $param_hash = $input{param};
  my $event_text = $event_ref->{text};
  my $event_link = $event_ref->{link};
  # Parse the event text for keywords.
  $event_text = $self->ParseText(text=>$event_text,
				 param=>$param_hash);
  # Parse the event link for keywords.
  $event_link = $self->ParseText(text=>$event_link,
				 param=>$param_hash);
  if(defined $event_link && $event_text =~ /VARLINK\(.*?\)/) {
    my $linkprefix;
    # Check if we need to add http:// to the beginning of the link. Any kind of
    # link text that already has the protocol is left alone. If you need
    # to link to ftp.somesite.com, then you need to include ftp:// in the link
    if($event_link =~ m|^.+?://|) {
      $linkprefix = qq{<a href="$event_link">};
    } else {
      $linkprefix = qq{<a href="http://$event_link">};
    }
    my $linksuffix = qq{</a>};
    $event_text =~ s/VARLINK\((.*?)\)/$linkprefix$1$linksuffix/g;
  }
  $event_ref->{text} = $event_text;
}

################################################################
# Fetch an array of events of a particular type
sub GetEvents {
  my $self = shift;
  my %hash = @_;
  my @events;
  # Figure out what kind of event the user wants.
  if(defined $hash{week}) {
    if(! defined $hash{scope}) {
      $hash{scope} = "week";
      $hash{scopevalue} = $hash{week};
    }
  }
  if(defined $hash{day}) {
    if(! defined $hash{scope}) {
      $hash{scope} = "day";
      $hash{scopevalue} = $hash{day};
    }
  }
  if(defined $hash{month}) {
    if(! defined $hash{scope}) {
      $hash{scope} = "month";
    }
  }
  foreach my $event (@{$self->get_events}) {
    if($event->{scope} eq "day" &&
       defined $hash{month} &&
       $self->get_month != $hash{month}) {
      next;
    }
    if($event->{scope} eq $hash{scope} &&
       $event->{scopevalue} eq $hash{scopevalue}) {
      if(defined $hash{type}) {
	if($event->{type} eq $hash{type}) {
	  push(@events,$event);
	}
      } else {
	if(! defined $event->{type}) {
	  push(@events,$event);
	}
      }
    }
  }
  if(! @events) {
    return ();
  }
  my @sorted_events =
      sort {
	($a->{intsortvalue} <=> $b->{intsortvalue}) ||
	    ($a->{alphsortvalue} cmp $b->{alphsortvalue})
	      } @events;
  return @sorted_events;
}

################################################################
# Some data initializer routines for objects that are hashes
sub show_weekcol {
  my $self = shift;
  my $pos = shift || 0;
  $self->get_columnopts->{weekcol} = 1;
  $self->get_columnopts->{weekcolpos} = $pos;
}
sub hide_weekcol {
  my $self = shift;
  $self->get_columnopts->{weekcol} = 0;
}

sub set_colourscheme {
  my $self = shift;
  my $scheme = shift || "prodblue";
  $self->{_colourscheme} = $scheme;
  $self->InitColours();
}

#
# Set the column options
#
# set_columnopts($hash_ref)
# set_columnopts(%hash)
#
sub set_columnopts {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_columnopts->{$option} = $param{$option};
  }
}

#
# Set calendar options
#
#
#
#
sub set_calendaropts {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_calendaropts->{$option} = $param{$option};
  }
}
sub set_rowopts {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_rowopts->{$option} = $param{$option};
  }
}
sub set_cellopts {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_cellopts->{$option} = $param{$option};
  }
}
sub set_VARS {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_VARS->{uc $option} = $param{$option};
  }
}

sub set_HTML {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_HTML->{$option} = $param{$option};
  }
}

sub set_tableopts {
  my $self = shift;
  my %param;
  if(ref($_[0]) eq "HASH") {
    %param = %{$_[0]};
  } else {
    %param = @_;
  }
  foreach my $option (keys %param) {
    $self->get_tableopts->{$option} = $param{$option};
  }
}
# hash data initializers
################################################################

################################################################
# Set the month of the calendar, filtering out bad values
sub set_month {
  my $self   = shift;
  my $month  = shift || $MKalendar::NOWMONTH;
  if($month < 1 || $month > 12) {
    $self->{_month} = $MKalendar::NOWMONTH;
  } else {
    $self->{_month} = $month;
  }
}
################################################################

################################################################
# Add a link from the day number in a day box.
#
# AddDayLink(day=>15,url=>"www.google.com");
#
#
sub AddDayLink {
  my $self = shift;
  my %hash = @_;
  my $linkurl = $hash{text} || $hash{link} || $hash{url};
  if(! defined $hash{day} || ! $linkurl) {
    return undef;
  }
  $self->get_daylinks->{$hash{day}} = $linkurl;
  return $linkurl;
}

################################################################
# Add an event to the calendar
#
# AddEvent(key=>value)
#
# key      value
# 
# day      day of the month (if event scope=day)
# week     week of the event  (if event scope=week)
#
# if either day or week is not defined, the event scope is
# taken as month
#
# intsortvalue   numberical sort value for event ordering
# alphsortvalue  alphabetical sort value for event ordering
sub AddEvent {
  my $self = shift;
  my %hash = @_;
  if(! defined $hash{intsortvalue}) {
    $hash{intsortvalue} = 0;
  }
  if(! defined $hash{alphsortvalue}) {
    $hash{alphsortvalue} = "";
  }
  if(defined $hash{day}) {
    # Event on a given day (e.g. 12)
    my $day = $hash{day};
    $hash{scope} = "day";
    $hash{scopevalue} = $hash{day};
    push(@{$self->{_events}},\%hash);
  } elsif (defined $hash{week}) {
    # Event on a given week (e.g. 2)
    my $week = $hash{week};
    $hash{scope} = "week";
    $hash{scopevalue} = $hash{week};
    push(@{$self->{_events}},\%hash);
  } else {
    # Event in the month (current month)
    $hash{scope} = "month";
    $hash{scopevalue} = $hash{$self->get_month};
    push(@{$self->{_events}},\%hash);
  }
}
#
################################################################

# This modifies the Date::Calc function
# and assigns sunday=0 rather than sunday=7.
sub DayofWeek {
  my @date = @_;
  my $dow = Day_of_Week(@date);
  if($dow == 7) {
    $dow = 0;
  }
  return $dow;
}
sub DowToText {
  my $dow = shift;
  if(! $dow) {
    $dow = 7;
  }
  return Day_of_Week_to_Text($dow);
}

sub DecodeDow {
  my %input = @_;
  my $dowtext = $input{dowtext} || return;
  my $decode = Decode_Day_of_Week($dowtext);
  if($decode == 7) {
    return 0;
  } 
  return $decode;  
}

sub VisibleDow {
  my %input = @_;
  my $dow = $input{dow};
  my $list  = $input{invlist};
  my $found = grep { $_ == $dow } @$list;
  return ! $found;
}

sub ColumnDow {
  my %input = @_;
  return ($input{cell} + $input{startdow}) % 7;
}

################################################################
# Construct the viewarray and viewheader objects
################################################################
sub Construct {
  my $self            = shift;
  my $month           = $self->get_month;
  my $year            = $self->get_year;

  $self->set_html(1);

  ################################################################
  # View initialization
  ################################################################

  my $startdow      = DecodeDow(dowtext=>$self->get_tableopts->{startdow});

  my @invisible_dow = map {DecodeDow(dowtext=>$_)}
    @{$self->get_tableopts->{disabledow}};
  my $show_week_of  = $self->get_tableopts->{showweekof};

  my $first_dow_of_month = DayofWeek($year,$month,1);

  ### Header construction [start]
  # The calendar columns each have a header. Most commonly, and by
  # default this is the first three letters of the day of the week
  my $column_idx=0;
  my $dow_column_idx=0;
  foreach my $cellnum (0..7) {
    my $cell_visible;
    $self->get_viewheaders->[$cellnum]->{col}     = $column_idx;
    if($cellnum == $self->get_columnopts->{weekcolpos}) {
      $cell_visible = $self->get_columnopts->{weekcol};
      $self->get_viewheaders->[$cellnum]->{week}  = 1;
    } else {
      my $header_dow = ColumnDow(cell=>$dow_column_idx,startdow=>$startdow);
      $cell_visible = VisibleDow(dow=>$header_dow,
				 invlist=>\@invisible_dow); 

      $self->get_viewheaders->[$cellnum]->{dow}     = $header_dow;
      $dow_column_idx ++;
    }
    $self->get_viewheaders->[$cellnum]->{visible} = $cell_visible;
    if($cell_visible) {
      $column_idx ++;
    }
  }
  ### Header construction [end]

  my @date_firstcell;
  my @date_cell;
  my $show_week = 1;
  my $row = 0;
  my $col = 0;
  foreach my $cellnum (0..50) {
#   $self->PrintDebug("cellnum $cellnum\n");
    # this is the first cell of the calendar. figure out the difference between the
    # first dow in the calendar and the first down of the month. @date holds the
    # date of the current cell
    if(! @date_firstcell) {
      # Calculate the difference (in days) between the first cell of
      # the calendar and the first day of the current month.
      my $diffdays = $first_dow_of_month - $startdow;
      if($diffdays < 0) {
	$diffdays += 7;
      }
      @date_firstcell = Add_Delta_Days($year,$month,1,-$diffdays);
      @date_cell = @date_firstcell;
    } else {
      # Advance the date by one.
      @date_cell = Add_Delta_Days(@date_firstcell,$cellnum);
    }
    # Check if we are on a new row in the next month
    if(not ($cellnum % 7)) {
      # Bug 08 Jan 2002
      # The year must be included in the comparison. The way it was before,
      # display of January which included December days broke because the
      # previous month (dec) was not greater than current month (jan)
      my $cellyear = $date_cell[0];
      my $cellmonth = $date_cell[1];
      if($cellyear+$cellmonth/100 > $self->get_year + $self->get_month/100) {
#	$self->PrintDebug("Celldate ",join(":",@date_cell),"\n");
#	$self->PrintDebug("Leaving cellnum loop\n");
	last;
      }
    }
    # Check if we should be showing this week.
    if(defined $show_week_of && not ($cellnum % 7) ) {
      if($show_week_of < 0 || $show_week_of > Days_in_Month($year,$month)) {
	$show_week = 1;
      } else {
	my $dd = Delta_Days(@date_cell,$year,$month,$show_week_of);
	if($dd < 0 || $dd >= 7) {
	  $show_week = 0;
	} else {
	  $show_week = 1;
	}
      }
    }
    my $cell_visible = 0;
    if($show_week &&
       VisibleDow(dow=>DayofWeek(@date_cell),invlist=>\@invisible_dow)) {
      $cell_visible = 1;
    }
    if(! defined $self->get_viewarray->[$row]) {
      $self->get_viewarray->[$row] = [];
    }
    $self->get_viewarray->[$row]->[$col] = {
					    dow=>DayofWeek(@date_cell),
					    date=>[@date_cell],
					    header=>$self->get_viewheaders->[$row],
					    visible=>$cell_visible
					   };
    if(not ((1+$cellnum) % 7)) {
      $col = 0;
      $row ++;
    } else {
      $col ++;
    }
  }
}


sub PrintCalendarSectionList {
  my $self = shift;
  my %input = @_;
  my $list = $input{list};
  my $param = $input{param};
  foreach my $section (@$list) {
    my $value;
    foreach my $modifier (@{$MKalendar::MODIFIERS}) {
      my $sectionname = lc $section . "_$modifier";
      if($self->get_VARS->{MODIFIER}->{$modifier} && defined $self->get_HTML->{$sectionname}) {
	$value = $self->get_HTML->{$sectionname};
	last;
      }
    }
    if(! defined $value) {
      $value = $self->get_HTML->{$section};
    }
    $HTML .= $self->ParseText(text=>$value,param=>$param) . "\n";
#    print $self->ParseText(text=>$value,
#			   param=>$param);
#    print "\n";
  }
}

# Alternate print function. More features!!!
#

sub Create {

  my $self = shift;
  my @SECTIONS;
  
  $HTML = "";

  @SECTIONS = qw(preamble
		 calendar_s
		 calendarheader_init
		 calendarheader_s
		 calendarheader
		 calendarheader_f
		 calendarheader_end
		);

  $self->get_VARS->{VARCALWIDTH} = $self->get_tableopts->{size};
  $self->get_VARS->{VARMONTH} = $self->get_month;
  $self->get_VARS->{VARNCOLUMN}  = grep {$_} map { $_->{visible} } @{$self->get_viewheaders};
  $self->get_VARS->{VARNCOLUMN} += $self->get_columnopts->{weekcol};

  $self->get_VARS->{VARCOLUMNWIDTH} = (defined $self->get_columnopts->{size})?
    $self->get_columnopts->{size}:int($self->get_tableopts->{size} / $self->get_VARS->{VARNCOLUMN})."%";

  $self->get_VARS->{VARMONTHTEXT} = Month_to_Text($self->get_month);
  $self->get_VARS->{VARMONTHTEXTSHORT} = substr(Month_to_Text($self->get_month),0,3);
  $self->get_VARS->{VARYEAR} = $self->get_year;
  $self->get_VARS->{VARVERSION} = $MKalendar::VERSION;
  $self->get_VARS->{VARCVSTAG} = $MKalendar::CVSTAG;
  $self->get_VARS->{VARCVSDATE} = $MKalendar::CVSDATE;
  $self->get_VARS->{VARHOUR} = (Now)[0];
  $self->get_VARS->{VARMIN} = sprintf("%02d",(Now)[1]);
  $self->get_VARS->{VARSEC} = sprintf("%02d",(Now)[2]);

  $self->PrintCalendarSectionList(list=>\@SECTIONS,
				  param=>$self->get_VARS);
  
  @SECTIONS = qw(columnheader_init
		 columnheader_s
		 columnheader
		 columnheader_f
		 columnheader_end
		);

  $self->PrintCalendarSectionList(list=>[qw(columnheader_init)],
				  param=>$self->get_VARS);

  foreach my $header (@{$self->get_viewheaders}) {
    if($header->{visible}) {
      @SECTIONS = qw(columnheader_s
		     columnheader
		     columnheader_f);
      $self->get_VARS->{VARDOW}     = $header->{dow};
      $self->get_VARS->{VARCOLUMN}  = $header->{col};
      $self->get_VARS->{VARDOWTEXT} = DowToText($header->{dow});
      $self->get_VARS->{VARDOWTEXTSHORT} = substr(DowToText($header->{dow}),0,3);

      if($header->{week}) {
	$self->get_VARS->{MODIFIER}->{WEEKCOL} = 1;
      } else {
	$self->get_VARS->{MODIFIER}->{WEEKCOL} = 0;
      } 
      if(defined $header->{dow}) {
	if($header->{dow} == 0 || $header->{dow} == 6) {
	  $self->get_VARS->{MODIFIER}->{WEEKEND} = 1;
	} else {
	  $self->get_VARS->{MODIFIER}->{WEEKEND} = 0;
	}
	if($header->{dow} == DayofWeek(Today)) {
	  $self->get_VARS->{MODIFIER}->{TODAYCOLUMN} = 1;
	} else {
	  $self->get_VARS->{MODIFIER}->{TODAYCOLUMN} = 0;
	}
      }
      $self->PrintCalendarSectionList(list=>\@SECTIONS,
				      param=>$self->get_VARS);
    }
  }
  $self->get_VARS->{MODIFIER}->{WEEKEND} = 0;
  $self->get_VARS->{MODIFIER}->{TODAYCOLUMN} = 0;
  $self->PrintCalendarSectionList(list=>[qw(columnheader_end)],
				  param=>$self->get_VARS);

  foreach my $row (0..@{$self->get_viewarray}-1) {
    $self->get_VARS->{VARWEEK} = $row+1;
    #
    # week row
    #
    if(! $row) {
      $self->PrintCalendarSectionList(list=>[qw(row_s)],
				      param=>$self->get_VARS);
    }
    my $calendar_row = $self->get_viewarray->[$row];
    my $visible_row_cells = 0;
    # Check if this row contains today's date.
    my $rowdate = $self->get_viewarray->[$row]->[0]->{date};
    if(Delta_Days(@$rowdate,Today) < 7 && Delta_Days(@$rowdate,Today) >=0 ) {
      $self->get_VARS->{MODIFIER}->{THISWEEK} = 1;
    } else {
      $self->get_VARS->{MODIFIER}->{THISWEEK} = 0;
    }
    my $calendar_columns = @{$calendar_row};
    foreach my $col (0..$calendar_columns-1) {
      #
      # week column
      #
      if($self->get_viewheaders->[$col]->{week} &&
	 $self->get_viewheaders->[$col]->{visible}) {
	$self->get_VARS->{MODIFIER}->{WEEKCOL} = 1;
	@SECTIONS = qw(
		       cell_init
		       cell_s
		      );
	$self->PrintCalendarSectionList(list=>\@SECTIONS,
					param=>$self->get_VARS);
	#
	# week events
	#

	my @week_events = $self->GetEvents(week=>$row+1);
	if(@week_events) {
	  $self->PrintCalendarSectionList(list=>[qw(event_init)],
					  param=>$self->get_VARS);
	  $self->get_VARS->{MODIFIER}->{WEEKEVENT} = 1;
	}
	foreach my $event_ref (@week_events) {
	  $self->ParseEvent(event=>$event_ref,
			    param=>$self->get_VARS);
	  $self->PrintCalendarSectionList(list=>[qw(event_s)],
					  param=>$self->get_VARS);
	  $HTML .= $event_ref->{text};
	  $self->PrintCalendarSectionList(list=>[qw(event_f)],
					  param=>$self->get_VARS);
	}
	if(@week_events) {
	  $self->PrintCalendarSectionList(list=>[qw(event_end)],
					  param=>$self->get_VARS);
	  $self->get_VARS->{MODIFIER}->{WEEKEVENT} = 0;
	}
	
	@SECTIONS = qw(
		       cell_f
		       cell_end
		      );
	$self->PrintCalendarSectionList(list=>\@SECTIONS,
					param=>$self->get_VARS);
      }
      $self->get_VARS->{MODIFIER}->{WEEKCOL} = 0;
      #
      # day columns
      #
      my $calendar_cell = $calendar_row->[$col];
      if(! $calendar_cell->{visible}) {
	next;
      }

      my $date = $calendar_cell->{"date"};
      
      $self->get_VARS->{MONTH} = $date->[1];
      $self->get_VARS->{VARMONTHTEXT} = Month_to_Text($date->[1]);
      $self->get_VARS->{VARMONTHTEXTSHORT} = substr(Month_to_Text($date->[1]),0,3);

      if($date->[1] != $self->get_month) {
	$self->get_VARS->{MODIFIER}->{OTHERMONTH} = 1;
      } else {
	$self->get_VARS->{MODIFIER}->{OTHERMONTH} = 0;
      }

      if($calendar_cell->{dow} == DayofWeek(Today)) {
	$self->get_VARS->{MODIFIER}->{TODAYCOLUMN} = 1;
      } else {
	$self->get_VARS->{MODIFIER}->{TODAYCOLUMN} = 0;
      }
      if($calendar_cell->{dow} == 0 || $calendar_cell->{dow} == 6) {
	$self->get_VARS->{MODIFIER}->{WEEKEND} = 1;
      } else {
	$self->get_VARS->{MODIFIER}->{WEEKEND} = 0;
      }
      #
      # TODAY is set if the column is a week column but WEEKCOL overrides TODAY
      #
      if(Delta_Days(@$date,Today) == 0) {
	$self->get_VARS->{MODIFIER}->{TODAY} = 1;
      } else {
	$self->get_VARS->{MODIFIER}->{TODAY} = 0;
      } 

      $self->get_VARS->{VARWEEKOFMONTH} = Week_Number(@$date);
      $self->get_VARS->{VARDAY}      = $date->[2];
      $self->get_VARS->{VARDAYDELTA} = Delta_Days(Today,@$date);
      $self->get_VARS->{VARDOW}      = DayofWeek(@$date);
      $self->get_VARS->{VARDOWTEXT}  = DowToText(DayofWeek(@$date));
      $self->get_VARS->{VARDOWTEXTSHORT} = substr(DowToText(DayofWeek(@$date)),0,3);
#     $self->get_VARS->{VARDOWNTH};
      $self->get_VARS->{VARCOLUMN}   = $self->get_viewheaders->[$col]->{col};

      @SECTIONS = qw(
		     cell_init
		     cell_s
		     cell_dayheader_s
		     cell_dayheader
		     cell_dayheader_f
		    );
      $self->PrintCalendarSectionList(list=>\@SECTIONS,
				      param=>$self->get_VARS);

      #
      # Check if there are any events for this day
      #
      my @day_events = $self->GetEvents(day=>$date->[2],month=>$date->[1]);
      if(@day_events) {
	$self->PrintCalendarSectionList(list=>[qw(event_init)],
					param=>$self->get_VARS);
	$self->get_VARS->{MODIFIER}->{DAYEVENT} = 1;
      }
      foreach my $event_ref (@day_events) {
	$self->ParseEvent(event=>$event_ref,
			  param=>$self->get_VARS);
	$self->PrintCalendarSectionList(list=>[qw(event_s)],
					param=>$self->get_VARS);
	$HTML .= $event_ref->{text};
	$self->PrintCalendarSectionList(list=>[qw(event_f)],
					param=>$self->get_VARS);
      }
      if(@day_events) {
	$self->PrintCalendarSectionList(list=>[qw(event_end)],
					param=>$self->get_VARS);
	$self->get_VARS->{MODIFIER}->{DAYEVENT} = 0;
      }
      @SECTIONS = qw(
		     cell_f
		     cell_end
		    );
     $self->PrintCalendarSectionList(list=>\@SECTIONS,
				      param=>$self->get_VARS);


      $visible_row_cells ++;
    }
    if($visible_row_cells) {
      $self->PrintCalendarSectionList(list=>[qw(row_f)],
				      param=>$self->get_VARS);
    }
  }
  $self->get_VARS->{VARBENCH} = sprintf("%.3f",Time::HiRes::tv_interval($MKalendar::EXECSTART));

  @SECTIONS = qw(
		 footer_init
		 footer_s
		);
  $self->PrintCalendarSectionList(list=>\@SECTIONS,
				  param=>$self->get_VARS);

  my @month_events = $self->GetEvents(month=>$self->get_month);
  if(@month_events) {
    $self->PrintCalendarSectionList(list=>[qw(event_init)],
				    param=>$self->get_VARS);
  }
  $self->get_VARS->{MODIFIER}->{MONTHEVENT} = 1;
  foreach my $event_ref (@month_events) {
    $self->ParseEvent(event=>$event_ref,
		      param=>$self->get_VARS);
    $self->PrintCalendarSectionList(list=>[qw(event_s)],
				    param=>$self->get_VARS);
    $HTML .= $event_ref->{text};
    $self->PrintCalendarSectionList(list=>[qw(event_f)],
				    param=>$self->get_VARS);
  }
  if(@month_events) {
    $self->PrintCalendarSectionList(list=>[qw(event_end)],
				    param=>$self->get_VARS);
  }

  @SECTIONS = qw(
		 footer
		 footer_f
		 footer_end);
  $self->PrintCalendarSectionList(list=>\@SECTIONS,
				  param=>$self->get_VARS);

  @SECTIONS = qw(stamp_init
		 stamp_s
		 stamp
		 stamp_f
		 stamp_end);

  $self->PrintCalendarSectionList(list=>\@SECTIONS,
				  param=>$self->get_VARS);

  @SECTIONS = qw(calendar_f
		 postamble);
  $self->PrintCalendarSectionList(list=>\@SECTIONS,
				  param=>$self->get_VARS);

  return $HTML;

}


################################################################
# Print the calendar
# sub Print {

#   my $self            = shift;
#   my $curr_month      = $self->get_month;
#   my $curr_month_text = Month_to_Text($curr_month);
#   my $curr_year       = $self->get_year;
#   my $days_in_month   = Days_in_Month($curr_year,$curr_month);

#   ################################################################
#   # CALENDAR TABLE DEFINITION AND HEADERS
#   #
#   ### table header showing month & year
#   #
#   my $table_width_type = $self->get_tableopts->{sizeopts};
#   my $table_width_size = $self->get_tableopts->{size};
#   my $table_width;
#   if($table_width_type =~ /rel/i) {
#     if(defined $table_width_size) {
#       $table_width = $table_width_size."\%";
#     }
#   }
#   if($table_width_type =~ /abs/i) {
#     if(defined $table_width_size) {
#       $table_width = "$table_width_size";
#     }
#   }

#   print "\n\n<!-- Calendar.pm $calendar::VERSION $calendar::CVSDATE -->\n\n";

#   ### Calendar table definition
#   #
#   # figure out how many columns we should be displaying.
#   #
#   # by default, all columns are enabled. to disable columns, use
#   #
#   # set_tableopts(disable_column=>[n1,n2,...]);
#   #
#   # where nx is the number of the column to disable. nx=-1 for the week column,
#   # nx=0 for the first dow in the calendar, nx=1 for the second, nx=6 for the last dow

#   my $disable_columns_list = $self->get_tableopts->{"disable_column"};
#   my $calendar_columns;
#   if($disable_columns_list) {
#     $calendar_columns = 8 - @{$disable_columns_list};
#   } else {
#     $calendar_columns = 8;
#   }

#   print "<table width=$table_width cellspacing=0 cellpadding=1 border=1>\n";
#   #
#   ################################################################
#   #
#   ### Print the month header
#   #
#   my $c_month_header = $self->get_rowopts->{monthcolour};
#   print
#       "<tr>",
#       "<td colspan=$calendar_columns align=center class=$c_month_header>",
#       "<b>",
#       $curr_month_text,
#       " ",
#       $curr_year,
#       "</b>",
#       "</td>",
#       "</tr>\n";
#   #
#   ### Print the day of week row
#   #
#   my $column_width_type = $self->get_columnopts->{sizeopts};
#   my $column_width_size = $self->get_columnopts->{size};
#   my $column_align      = $self->get_columnopts->{align};
#   my $column_width;
#   if($column_width_type =~ /rel/) {
#     if(defined $column_width_size) {
#       $column_width = "$column_width_size%";
#     }
#   }
#   if($column_width_type =~ /abs/) {
#     if(defined $column_width_size) {
#       $column_width = "$column_width_size";
#     }
#   }
#   print "<tr>\n";
#   #
#   # The first column is the "week" column. This corresponds to dow_idx -1.
#   #
#   for(my $dow_idx=-1;$dow_idx<7;$dow_idx++) {
#     # Check if we should skip this column.
#     if(grep {$_ == $dow_idx} @$disable_columns_list) {
#       next;
#     }
#     my $dow_num       = ($dow_idx-1)%7+1;
#     my $dow_text      = Day_of_Week_to_Text($dow_num);
#     my $dow_text_abbv = substr($dow_text,0,3);
#     my $dow_cell_colour;
#     my $week_header_colour = $self->get_rowopts->{weekheadcolour};
#     if($dow_num < 6) {
#       $dow_cell_colour = $self->get_rowopts->{dowcolour};
#     } else {
#       $dow_cell_colour = $self->get_rowopts->{weekendcolour};
#     }
#     if($dow_idx == -1) {
#       $dow_text = $dow_text_abbv = "Week";
#       $dow_cell_colour = $week_header_colour;
#     }
#     print 
# 	"<td width=$column_width class=$dow_cell_colour align=$column_align>",
# 	$dow_text_abbv,
# 	"</td>\n";
#   }
#   print "</tr>\n\n";
#   #
#   ################################################################
#   #
#   ### Now print the calendar body - a row for each week. The iteration
#   ### is done by week rather than by day.
#   #
#   my $week_of_month_idx = 1;
#   my $day_idx = 1;
#   my %dow_nth;

#   # Calculate the # of weeks in the month
#   my $first_day_dow_num = (Day_of_Week($curr_year,$curr_month,$day_idx) % 7) + 1;
#   my $last_day_dow_num   = (Day_of_Week($curr_year,$curr_month,$days_in_month) % 7) + 1;
#   my $weeks_in_month    = POSIX::ceil((7-$first_day_dow_num +
# 			   + $days_in_month
# 			   + 7-$last_day_dow_num)/7);

#   my %calendar_cell_param;
#   $calendar_cell_param{VARMONTH} = $curr_month;
#   $calendar_cell_param{VARMONTHTEXT} = $curr_month_text;
#   $calendar_cell_param{VARYEAR} = $curr_year;

#   for(my $week_idx=1;$week_idx<=$weeks_in_month;$week_idx++) {
#     $calendar_cell_param{VARWEEKOFMONTH} = $week_idx;
#     print "<!-- Week number $week_idx of $curr_month-$curr_year -->\n\n";
#     print "<tr>\n";
#     #
#     # First print the week column cell
#     #
#     my $weeknumcolour = $self->get_rowopts->{weeknumcolour};
#     my $weekcellcolour = $self->get_rowopts->{weekcellcolour};
#     if(! grep(/\-1/,@$disable_columns_list)) {
#       print 
# 	"<td valign=top class=$weekcellcolour>",
# 	  "<div align=center class=$weeknumcolour>",
# 	    "<span class=small>$week_idx</span>",
# 	      "</div>\n";
      
#       #
#       # Collect events for this week.
#       #
#       my @week_events = $self->GetEvents(week=>$week_idx);
#       foreach my $event_ref (@week_events) {
# 	$self->ParseEvent(event=>$event_ref,
# 			  param=>\%calendar_cell_param);
# 	$self->PrintEvent($event_ref);
#       }
#       print
# 	"</td>\n\n";
#     }
#     # Iterate through the days until we reach the last day of this week.
#     # This loop quits when we get to the last day of the week
#     do {
#       my $dow_num  = Day_of_Week($curr_year,$curr_month,$day_idx);
#       my $dow_text = Day_of_Week_to_Text($dow_num);
#       my $col_idx  = ($dow_num % 7)+1;
#       # Index of the day of the week (e.g. 2nd tuesday)
#       $dow_nth{$dow_text}++;
#       my $dow_nth  = $dow_nth{$dow_text};

#       $calendar_cell_param{VARDAY} = $day_idx;
#       $calendar_cell_param{VARDAYDELTA} = Delta_Days($calendar::NOWYEAR,
# 						     $calendar::NOWMONTH,
# 						     $calendar::NOWDAY,
# 						     $curr_year,
# 						     $curr_month,
# 						     $day_idx);
#       $calendar_cell_param{VARDOW} = $dow_num;
#       $calendar_cell_param{VARDOWTEXT} = $dow_text;
#       $calendar_cell_param{VARCOLUMN} = $col_idx;
#       $calendar_cell_param{VARDOWNTH} = $dow_nth;

#       my $c_blank_cell = $self->get_cellopts->{blankcolour};
#       my $c_today_cell = $self->get_cellopts->{todaycolour};
#       my $c_weekend_cell = $self->get_cellopts->{weekendcolour};

#       print "<!-- Day $day_idx-$curr_month-$curr_year ($dow_text) -->\n";

#       #
#       # If this is the first day of the month, skip some cells.
#       #
#       if($day_idx == 1) {
# 	# Print the week column
# 	if(my $colspan_width = $col_idx-1) {
# 	  print "<td align=right colspan=$colspan_width class=$c_blank_cell></td>\n";
# 	}
#       }
#       my $day_box_color = $self->get_cellopts->{colour};
#       # If this is a weekend, then change the colour
#       if($dow_num == 6 || $dow_num == 7) {
# 	$day_box_color  = $self->get_cellopts->{weekendcolour};
#       }
#       # If this is the current week, then change the colour
#       if(
# 	 Week_Number($curr_year,
# 		     $curr_month,
# 		     ($day_idx<$days_in_month)?$day_idx+1:$day_idx
# 		     ) ==
# 	 Week_Number($calendar::NOWYEAR,
# 		     $calendar::NOWMONTH,
# 		     $calendar::NOWDAY)
# 	 ) {
# 	$day_box_color  = $self->get_cellopts->{thisweekcolour};
#       }
#       if($self->IsDateToday($curr_year,$curr_month,$day_idx)) {
# 	$day_box_color  = $self->get_cellopts->{todaycolour};
#       }
#       # Start the day cell
#       print 
# 	  "<td valign=top class=$day_box_color>\n";
#       my $day_number_header = $self->ParseText(text=>$self->get_cellopts->{daynumberheader},
# 					       param=>\%calendar_cell_param);
#       print 
# 	  "<div valign=top align=left>";
#       if(my $daylink = $self->get_daylinks->{$day_idx}) {
# 	print qq{<a href="http://$daylink">$day_number_header</a>};
#       } else {
# 	print $day_number_header;
#       }
#       print
# 	  "\n";
#       # Now print the header.
#       my @day_headers = $self->GetEvents(day=>$day_idx,type=>"header");
#       foreach my $event_ref (@day_headers) {
# 	$self->ParseEvent(event=>$event_ref,
# 			  param=>\%calendar_cell_param);
# 	print $event_ref->{text};
#       }

#       print "</div>";
#       # Check if there are any events for this day
#       my @day_events = $self->GetEvents(day=>$day_idx);
#       foreach my $event_ref (@day_events) {
# 	$self->ParseEvent(event=>$event_ref,
# 			  param=>\%calendar_cell_param);
# 	print $event_ref->{text};
#       }
#       print
# 	  "</td>\n";
#       # Handle the last day of the month
#       if($day_idx == $days_in_month) {
# 	if(my $colspan_width = 7-$col_idx) {
# 	  print "<td colspan=$colspan_width class=$c_blank_cell></td>\n";
# 	}
# 	print "</tr>\n";
#       }
#       $day_idx ++;
#       if($day_idx > $days_in_month) {last}
#     } while (Day_of_Week($curr_year,$curr_month,$day_idx) <= 6);
#     print "</tr>\n";
#   }
#   print "</tr>\n\n";

#   my @month_events = $self->GetEvents(month=>$curr_month);
#   if(@month_events) {
#     print "<tr>\n";
#     print "<td colspan=8>\n";
#     foreach my $event_ref (@month_events) {
#       $self->ParseEvent(event=>$event_ref,
# 			param=>\%calendar_cell_param);
#       $self->PrintEvent($event_ref);
#     }
#     print "</td>\n";
#     print "</tr>\n";
#   }

#   print 
#       "<tr>",
#       "<td colspan=8 align=right>";
#   $self->VersionBar($self->get_calendaropts);
#   print
#       "</td></tr>";
#   print
#       "</table>";

# }
################################################################

################################################################
# Check if the passed date is today
sub IsDateToday {
  my $self = shift;
  my @date = @_;
  if($date[0] == $calendar::NOWYEAR &&
     $date[1] == $calendar::NOWMONTH &&
     $date[2] == $MKalendar::NOWDAY) {
    return 1;
  } else {
    return 0;
  }
}
################################################################

################################################################
# Print the contents of the calendar data structure
sub DataDump {
  my $self = shift;
  use Data::Dumper;
  my $dumpdata = Dumper($self);
  $self->ScreenPrint($dumpdata);
}
################################################################

sub PrintEvent {
  my $self = shift;
  my $event_ref = shift;
  my $event_text = $event_ref->{text};
  my $event_link = $event_ref->{link};
  if(defined $event_link && $event_text =~ /VARLINK\(.*?\)/) {
    my $linkprefix = qq{<a href="http://$event_link">};
    my $linksuffix = qq{</a>};
    $event_text =~ s/VARLINK\((.*?)\)/$linkprefix$1$linksuffix/g;
  }
  print $event_text;
}

################################################################
# Debug code
sub Debug {
  my $self = shift;
  if(defined $_[0]) {
    $MKalendar::DEBUG = $_[0];
  }
  $MKalendar::DEBUG;
}
sub PrintDebug {
  my $self=shift;
  if($self->Debug) {
    $self->ScreenPrint(@_);
    print "\n";
  }
}
################################################################

################################################################
# Provides a proper output function. If the html mode is set
# then all newlines are replaced by <br>+newline.
sub ScreenPrint {
  my $self = shift;
  my $text = join("",@_);
  if($self->get_html) {
    $text =~ s/\n/<br>\n/g;
  }
  print $text;
}
################################################################

################################################################
# Automated functions. 
# Do not edit below this point, unless you know what you are
# doing.

sub MKalendar::DESTROY {
  my $self = shift;
}

sub MKalendar::AUTOLOAD {
  no strict "refs";
  my ($self,$newval) = @_;

  if ($AUTOLOAD =~ /.*::get(_\w+)/ && ($self->_accessible($1,'read'))) {
    my $attr_name = $1;
    *{$AUTOLOAD} = sub { return $_[0]->{$attr_name} };
    return $self->{$attr_name}
  }
  if ($AUTOLOAD =~ /.*::set(_\w+)/ && $self->_accessible($1,'write')) {
    my $attr_name = $1;
    *{$AUTOLOAD} = sub { $_[0]->{$attr_name} = $_[1]; return };
    if(defined $newval) {
      $self->{$1} = $newval;
    } else {
      $self->{$1} = 1;
    }
    return $self->{$1}
  }

  carp "No such method: $AUTOLOAD";

}
# END
################################################################

1;
